/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Gabber 2
 *  Based on Gabber, Copyright (c) 1999-2002 Dave Smith & Julian Missig
 *  Copyright (c) 2002-2003 Julian Missig
 */

#include "ContactInfo.hh"

#include "JabberConnection.hh"

#include "GabberUtility.hh"
#include "GabberWidgets.hh"

#include "intl.h"

#include <jabberoo/packet.hh>
#include <jabberoo/vCard.h>

#include <gtkmm/dialog.h>
#include <gtkmm/entry.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/sizegroup.h>
#include <gtkmm/textview.h>
#include <gtkmm/window.h>

using namespace jabberoo;

namespace Gabber {

// ---------------------------------------------------------
//
// Contact Information Dialog
//
// ---------------------------------------------------------

void ContactInfoDlg::display(JabberConnection& conn, Gtk::Window& parent,
			     const Glib::ustring& jid)
{
     ContactInfoDlg* e = manage(new ContactInfoDlg(conn, parent, jid));
     e->show();
}

ContactInfoDlg::ContactInfoDlg(JabberConnection& conn, Gtk::Window& parent,
			       const Glib::ustring& jid)
     : BaseGabberWindow("ViewInfo_dlg"), _conn(conn), _jid(jid)
{
     // Set the dialog's parent
     Gtk::Dialog* dlg  = static_cast<Gtk::Dialog*>(getGtkWindow());
     dlg->set_transient_for(parent);

     Gtk::Button* b;
     get_widget("Close_btn", b);
     b->signal_clicked().connect(slot(*this, &ContactInfoDlg::close));

     // Align the labels with the magic of SizeGroup
     Gtk::Label* l;
     Glib::RefPtr<Gtk::SizeGroup> client_grp = Gtk::SizeGroup::create(Gtk::SIZE_GROUP_HORIZONTAL);
     get_widget("JabberClientName_lbl", l);
     client_grp->add_widget(*l);
     get_widget("JabberClientVersion_lbl", l);
     client_grp->add_widget(*l);
     get_widget("ComputerOS_lbl");
     client_grp->add_widget(*l);
     get_widget("ComputerTime_lbl");
     client_grp->add_widget(*l);

     _presence_query = _conn.getSession().registerXPath(
        "/presence[@from='" + _jid + "']",
        slot(*this, &ContactInfoDlg::on_presence_node));
     
     _prettyjid = manage(new PrettyJID(_jid, "", PrettyJID::dtJIDRes, true));
     _prettyjid->show();
     _prettyjid->changed.connect(slot(*this, &ContactInfoDlg::on_PrettyJID_changed));
     on_PrettyJID_changed(); // Fire once since this won't normally be fired when we init

     Gtk::HBox* hb;
     get_widget("JIDInfo_hbox", hb);
     hb->pack_start(*_prettyjid, true, true, 0);
     _nickname = _prettyjid->get_nickname();

     // Set the window title
     _thisWindow->set_title(Util::substitute(_("%s's Info"), _nickname));

     Gtk::Entry* nickent;
     get_widget("Nickname_ent", nickent);
     nickent->set_text(_nickname);
     nickent->set_editable(_prettyjid->is_on_roster());

     // Set the s10n info
     try {
          // grab the roster item
          jabberoo::Roster::Item item = _conn.getSession().roster()[jabberoo::JID::getUserHost(_jid)];

          // set the s10n labels
          get_widget("S10nType_lbl_lbl", l);
          l->set_label(Util::getS10nName(item.getSubsType()) + ":");
          get_widget("S10nType_lbl", l);
          l->set_label(Util::getS10nInfo(item.getSubsType()));

          // display "Pending" if applicable
          if (item.isPending())
          {
               get_widget("S10nPending_lbl_lbl", l);
               l->show();
               get_widget("S10nPending_lbl", l);
               l->show();
          }

          // display "Subscribe" button if applicable
          switch (item.getSubsType())
          {
               case Roster::rsFrom:
               case Roster::rsNone:
                    get_widget("Subscribe_btn", b);
                    b->show();
                    get_widget("Subscribe_lbl", l);
                    l->show();
                    break;
               default:
                    break;
          }
          
     } catch (jabberoo::Roster::XCP_InvalidJID& e) {
          get_widget("AddContact_btn", b);
          b->show();
          get_widget("AddContact_lbl", l);
          l->show();
     }

     send_vcard_request();
}

void ContactInfoDlg::close()
{
     // Figure out if the nickname has changed, and if so, update it
     Gtk::Entry* nickent;
     get_widget("Nickname_ent", nickent);
     Glib::ustring newnick = nickent->get_text();
     
     // no empty nicks, thank you
     if (newnick.empty())
     {
          Gtk::MessageDialog dlg(*_thisWindow,
                                 _("Empty Display Names are not allowed. The Display Name has been changed back to its previous value."),
                                 Gtk::MESSAGE_WARNING, Gtk::BUTTONS_OK, true);
          dlg.set_title(_("Display Name cannot be empty"));
          dlg.run();
          nickent->set_text(_prettyjid->get_nickname());
          return;
     }
     
     if (newnick != _prettyjid->get_nickname())
     {
          try {
               jabberoo::Roster::Item item = _conn.getSession().roster()[jabberoo::JID::getUserHost(_jid)];

               // Set the new nickname
               item.setNickname(newnick);
               _conn.getSession().roster() << item;
               
          } catch (jabberoo::Roster::XCP_InvalidJID& e) {
               Gtk::MessageDialog dlg(*_thisWindow,
                                      Util::substitute(_("The Display Name cannot be changed because this Contact (%s) was not found on the Contact List. This is probably a bug in Gabber."), _jid),
                                      Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
               dlg.set_title(_("Contact cannot be found"));
               dlg.run();
               nickent->set_text(_prettyjid->get_nickname());
               return;
          }
     }

     // No more presence xpath for us
     _conn.getSession().unregisterXPath(_presence_query);
     
     // destroy the Gtk::Window
     BaseGabberWindow::close();
}

void ContactInfoDlg::send_vcard_request()
{
     // Get next session ID
     string id = _conn.getSession().getNextID();

     // Construct vCard request
     Packet iq("iq");
     iq.setID(id);
     iq.setTo(_jid);
     iq.getBaseElement().putAttrib("type", "get");
     judo::Element* vCard = iq.getBaseElement().addElement("vCard");
     vCard->putAttrib("xmlns", "vcard-temp");
     vCard->putAttrib("version", "2.0");
     vCard->putAttrib("prodid", "-//HandGen//NONSGML vGen v1.0//EN");

     // Send the vCard request
     _conn.getSession() << iq;
     _conn.getSession().registerIQ(id, slot(*this, &ContactInfoDlg::parse_vcard));
}

void ContactInfoDlg::set_info_label(const Glib::ustring& label_name, const Glib::ustring& data)
{
     if (data.empty())
	  return;

     Gtk::Label* l;
     get_widget(label_name + "_lbl", l);
     l->show();
     get_widget(label_name, l);
     l->set_label(data);
     l->show();
}

void ContactInfoDlg::parse_vcard(const judo::Element& t)
{
     if (t.empty())
	  return;

     const judo::Element* vCardElem = t.findElement("vCard");
     if (vCardElem == NULL)
	  return;

     Glib::ustring curtext;
 
     vCard vc(*vCardElem);

     // General
     curtext = vc[vCard::Fullname];
     set_info_label("FullName_lbl", curtext);
     curtext = vc[vCard::EMail];
     set_info_label("Email_lbl", curtext);
     curtext = vc[vCard::URL];
     set_info_label("WebSite_lbl", curtext);
     curtext = vc[vCard::Telephone];
     set_info_label("Telephone_lbl", curtext);

     // About
     curtext = vc[vCard::Birthday];
     if (!curtext.empty())
     {
	  Gtk::Widget* f;
	  get_widget("Birthday_frame", f);
	  f->show();
	  set_info_label("Birthday_lbl", curtext);
     }
     curtext = vc[vCard::Description];
     Gtk::TextView* tv;
     get_widget("About_txtview", tv);
     tv->get_buffer()->set_text(curtext);
}

void ContactInfoDlg::send_version_request()
{
     // Get next session ID
     string id = _conn.getSession().getNextID();

     // Construct version request
     Packet iq("iq");
     iq.setID(id);
     iq.setTo(_prettyjid->get_full_jid());
     iq.getBaseElement().putAttrib("type", "get");
     judo::Element* vCard = iq.getBaseElement().addElement("query");
     vCard->putAttrib("xmlns", "jabber:iq:version");

     // Send the version request
     _conn.getSession() << iq;
     _conn.getSession().registerIQ(id, slot(*this, &ContactInfoDlg::parse_version));
}

void ContactInfoDlg::parse_version(const judo::Element& t)
{
     if (t.empty())
	  return;

     const judo::Element* query = t.findElement("query");
     if (query == NULL)
	  return;
	 
     set_info_label("JabberClientName_lbl", query->getChildCData("name"));
     set_info_label("JabberClientVersion_lbl", query->getChildCData("version"));
     set_info_label("ComputerOS_lbl", query->getChildCData("os"));
}

void ContactInfoDlg::send_time_request()
{
     // Get next session ID
     string id = _conn.getSession().getNextID();

     // Construct version request
     Packet iq("iq");
     iq.setID(id);
     iq.setTo(_prettyjid->get_full_jid());
     iq.getBaseElement().putAttrib("type", "get");
     judo::Element* vCard = iq.getBaseElement().addElement("query");
     vCard->putAttrib("xmlns", "jabber:iq:time");

     // Send the version request
     _conn.getSession() << iq;
     _conn.getSession().registerIQ(id, slot(*this, &ContactInfoDlg::parse_time));
}

void ContactInfoDlg::parse_time(const judo::Element& t)
{
     if (t.empty())
	  return;

     const judo::Element* query = t.findElement("query");
     if (query == NULL)
	  return;

     Glib::ustring curtext;
     curtext = query->getChildCData("display");
     curtext += " "; // Don't duplicate the time zone:
     if(curtext.find(query->getChildCData("tz"))==string::npos)
	  curtext += query->getChildCData("tz");
     set_info_label("ComputerTime_lbl", curtext);
}

void ContactInfoDlg::get_status()
{
     // We want to catch a different presence this time
     _conn.getSession().unregisterXPath(_presence_query);
     _presence_query = _conn.getSession().registerXPath(
        "/presence[@from='" + _prettyjid->get_full_jid() + "']",
        slot(*this, &ContactInfoDlg::on_presence_node));
     
     // copy & paste from PrettyJID

     // Grab an existing presence for this particular resource (maybe)
     // and process it
     try {
          const jabberoo::Presence& p = _conn.getSession().presenceDB().findExact(_prettyjid->get_full_jid());
          on_presence_node(p.getBaseElement());
     } catch (jabberoo::PresenceDB::XCP_InvalidJID& e) {
          // Construct an offline presence
          jabberoo::Presence p("", jabberoo::Presence::ptUnavailable, jabberoo::Presence::stOffline, _("No presence has been received."));
          p.setFrom(_prettyjid->get_full_jid());
          on_presence_node(p.getBaseElement());
     }

     // end copy & paste from PrettyJID
}

void ContactInfoDlg::on_presence_node(const judo::Element& e)
{
     const Presence p(e);

     Gtk::Label* l;
     get_widget("Status_lbl", l);
     if (p.getStatus().empty())
          l->set_label(Util::getShowName(p.getShow()));
     else
          l->set_label(Util::substitute(_("%s: %s"), Util::getShowName(p.getShow()), p.getStatus()));

     // Whether or not the last request should be considered "offline"
     bool offline = false;
     if (p.getShow() == Presence::stOffline ||
         p.getShow() == Presence::stInvalid)
          offline = true;
     
     send_last_request(offline);
}

void ContactInfoDlg::send_last_request(bool offline)
{
     // Get next session ID
     string id = _conn.getSession().getNextID();

     // Construct version request
     Packet iq("iq");
     iq.setID(id);
     iq.setTo(_prettyjid->get_full_jid());
     iq.getBaseElement().putAttrib("type", "get");
     judo::Element* vCard = iq.getBaseElement().addElement("query");
     vCard->putAttrib("xmlns", "jabber:iq:last");

     // Send the version request
     _conn.getSession() << iq;
     _conn.getSession().registerIQ(id, SigC::bind(slot(*this, &ContactInfoDlg::parse_last), offline));
}

void ContactInfoDlg::parse_last(const judo::Element& t, bool offline)
{
     if (!t.empty())
     {
          int seconds;
          string lasttime;
          const judo::Element* query = t.findElement("query");
          if (query != NULL)
          {
               // Grab the idle or last logged out time
               seconds = atoi(query->getAttrib("seconds").c_str());
               if (seconds == 0)
                    return;
               
               // Conversion magic
               int mins, hrs, days, secs;
               days = (seconds / (3600 * 24));
               hrs  = ((seconds / 3600) - (days * 24));
               mins = ((seconds / 60) - (days * 24 * 60) - (hrs * 60));
               secs = (seconds - (days * 24 * 60 * 60) - (hrs * 60 * 60) - (mins * 60));
               char *tdays, *thrs, *tmins, *tsecs;
               tdays = g_strdup_printf("%d", days);
               thrs  = g_strdup_printf("%d", hrs);
               tmins = g_strdup_printf("%d", mins);
               tsecs = g_strdup_printf("%d", secs);

               // Figure out the exact text to display
               bool prevt;
               if (string(tdays) == "1")
               {
                    lasttime += " " + string(tdays) + _(" day");
                    prevt = true;
               }
               else if (string(tdays) != "0")
               {
                    lasttime += " " + string(tdays) + _(" days");
                    prevt = true;
               }
               else
                    prevt = false;
               g_free(tdays);

               if (string(thrs) == "1")
               {
                    if (prevt)
                         lasttime += ",";
                    lasttime += " " + string(thrs) + _(" hr");
                    prevt = true;
               }
               else if (string(thrs) != "0")
               {
                    if (prevt)
                         lasttime += ",";
                    lasttime += " " + string(thrs) + _(" hrs");
                    prevt = true;
               }
               else
                    prevt = false;
               g_free(thrs);

               if (string(tmins) == "1")
               {
                    if (prevt)
                         lasttime += ",";
                    lasttime += " " + string(tmins) + _(" min");
                    prevt = true;
               }
               else if (string(tmins) != "0")
               {
                    if (prevt)
                         lasttime += ",";
                    lasttime += " " + string(tmins) + _(" mins");
                    prevt = true;
               }
               else
                    prevt = false;
               g_free(tmins);

               if (!prevt)
               {
                    if (string(tsecs) == "1")
                         lasttime += " " + string(tsecs) + _(" sec");
                    else if (string(tsecs) != "0")
                         lasttime += " " + string(tsecs) + _(" secs");
               }
               g_free(tsecs);


               Gtk::Label* l;
               get_widget("Last_lbl", l);
               l->show();

               // last logged out or idle?
               if (offline)
               {
                    l->set_label(Util::substitute(_("Last logged out %s ago"), lasttime));

                    get_widget("Status_lbl", l);
                    l->set_label(Util::substitute(_("%s: %s"), _("Offline"), t.getChildCData("query"))); // Get the status from iq:last
               }
               else
               {
                    l->set_label(Util::substitute(_("Idle for %s"), lasttime));
               }
          }
     }     
}

void ContactInfoDlg::on_PrettyJID_changed()
{
     send_version_request();
     send_time_request();
     get_status();

     // Hide the idle labels.
     // get_status() will call on_presence(), which will send_last_request() for us
     Gtk::Label* l;
     get_widget("Last_lbl", l);
     l->hide();
}

} // namespace Gabber
