/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Gabber 2
 *  Based on Gabber, Copyright (c) 1999-2002 Dave Smith & Julian Missig
 *  Copyright (c) 2002 Julian Missig
 */

#include "GabberUtility.hh"
#include "PacketQueue.hh"

#include <jabberoo/JID.hh>
#include <jabberoo/filestream.hh>

#include <glibmm/convert.h>
#include <glibmm/fileutils.h>
#include <glibmm/miscutils.h>

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#include <functional>

using namespace Gabber;

PacketQueue::PacketQueue(jabberoo::Session& sess) : 
    _session(sess)
{
    // Make the directories if needed
    _gdir = Glib::build_filename(Glib::get_home_dir(),".Gabber");
    if (!Glib::file_test(_gdir, Glib::FILE_TEST_EXISTS | Glib::FILE_TEST_IS_DIR))
    {
        mkdir(_gdir.c_str(), 0700);
    }
    _gdir = Glib::build_filename(_gdir, "PacketQueue");
    if (!Glib::file_test(_gdir, Glib::FILE_TEST_EXISTS | Glib::FILE_TEST_IS_DIR))
    {
        mkdir(_gdir.c_str(), 0700);
    }
}

PacketQueue::~PacketQueue()
{
}

void PacketQueue::push(judo::Element* elem)
{
    elem->putAttrib("gabber:queued", "true");
    std::ofstream fil;
    std::string userhost =
        jabberoo::JID::getUserHost(elem->getAttrib("from"));
    Glib::ustring filename = Glib::build_filename(_gdir, userhost);
    fil.open(filename.c_str(), std::ios::app);
    fil << elem->toString();
    fil.close();

    packet_queued_event(userhost);

    return;
}

void PacketQueue::pop(const std::string& jid)
{
    if (!isQueued(jid))
        return;

    Glib::ustring filename = Glib::build_filename(_gdir, 
        jabberoo::JID::getUserHost(jid));
    jabberoo::FileStream fs(filename.c_str());
    if (fs.ParseFile("queue"))
    {
        judo::Element* root = fs.getRoot();
        if (root && root->size() > 0)
        {
            judo::Element::iterator it = root->begin();
            judo::Element* child =
                static_cast<judo::Element*>(root->detachChild(it));
            _session.dispatch(child);

            if (root->size() > 0)
            {
                std::ofstream fil;
                fil.open(filename.c_str());
                std::for_each(root->begin(), root->end(), saveToFile(fil));
                fil.close();

                packet_queued_event(jid);
            }
            else
            {
                unlink(filename.c_str());
                queue_emptied_event(jid);
            }

            delete root;
        }
        else if (root && root->size() == 0)
        {
            unlink(filename.c_str());
            queue_emptied_event(jid);
        }
    }
}

void PacketQueue::loadFromDisk()
{
    // Read through a directory and read in the files
    Glib::Dir dir(_gdir);

    // XXX More STLish?
    for(Glib::Dir::iterator it = dir.begin(); it != dir.end(); ++it)
    {
        std::string fname = Glib::build_filename(_gdir, (*it));
        packet_queued_event(fname);
    }
}

bool PacketQueue::isQueued(const std::string& jid)
{
    std::string fname = Glib::build_filename(_gdir,
        jabberoo::JID::getUserHost(jid));
    return Glib::file_test(fname, Glib::FILE_TEST_EXISTS);
}

judo::Element* PacketQueue::get(const std::string& jid)
{
    if (!isQueued(jid))
    {
        throw notQueued();
    }

    Glib::ustring filename = Glib::build_filename(_gdir, 
        jabberoo::JID::getUserHost(jid));
    jabberoo::FileStream fs(filename.c_str());
    if (fs.ParseFile("queue"))
    {
        return fs.getRoot();
    }
    else
    {
        throw invalidQueue();
    }
}

void PacketQueue::set(const std::string& jid, const judo::Element& queue)
{
    if (queue.getName() != "queue")
    {
        throw invalidQueue();
    }

    Glib::ustring filename = Glib::build_filename(_gdir, 
        jabberoo::JID::getUserHost(jid));

    std::ofstream fil;
    fil.open(filename.c_str());
    std::for_each(queue.begin(), queue.end(), saveToFile(fil));
    fil.close();

    packet_queued_event(jid);
}

PacketQueue::saveToFile::saveToFile(std::ofstream& ofs) : _ofs(ofs)
{ }

void PacketQueue::saveToFile::operator()(const judo::Node* node)
{
    if (node->getType() == judo::Node::ntElement)
    {
        _ofs << static_cast<const judo::Element*>(node)->toString();
    }
}
