/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Gabber 2
 *  Based on Gabber, Copyright (c) 1999-2002 Dave Smith & Julian Missig
 *  Copyright (c) 2002 Julian Missig
 */

#ifndef INCL_GABBER_PACKETQUEUE_HH
#define INCL_GABBER_PACKETQUEUE_HH

#include <jabberoo/session.hh>
#include <jabberoo/judo.hpp>
#include <jabberoo/packetqueue.hh>

#include <glibmm/ustring.h>

#include <fstream>
#include <string>
#include <list>
#include <map>

namespace Gabber{

/// Queues incoming packets based on jids.
class PacketQueue : public jabberoo::PacketQueue
{
public:
    // Exception structs
    struct notQueued { };
    struct invalidQueue { };

public:
    PacketQueue(jabberoo::Session& sess);
    ~PacketQueue();

    /**
    * Push the element on to the back of the packet list.
    * @param elem The element to add to the list.
    */
    void push(judo::Element* elem);

    /**
    * Pops the first packet from the queue, for the specified JID,
    * and hands it off to the PacketDispatcher.
    * @param jid The JID (only user@host used) to process for.
    */
    void pop(const std::string& jid);

    /**
    * Load the pending queue from the disk and fire events for it.
    * This can be called whenever a refresh of the pending events is needed.
    */
    void loadFromDisk();

    /**
    * See if the specified JID has queued packets.
    * @param jid The user@host jid to check.
    * @returns bool true if there are packets queued for the jid.
    */
    bool isQueued(const std::string& jid);

    /**
    * Get the queue for the specified JID.
    * The queue may be manipulated without affecting what is on disk.  To
    * actually save the changes after manipulation use the set call.  The caller
    * is responsible for freeing the resulting queue.
    * @param jid The user@host jid to check.
    * @returns judo::Element& The root queue element
    */
    judo::Element* get(const std::string& jid);

    /**
    * Set the queue for the specified JID and save it to disk.
    * This will remove what is already on the disk, if you wish to add to what
    * is already present, use push instead.
    * @param jid The user@host jid to save to.
    * @param queue The entire <queue/> element to save.
    */
    void set(const std::string& jid, const judo::Element& queue);

    /// Event that is fired whenever a packet is queued for the specified user.
    SigC::Signal1<void, const std::string&> packet_queued_event;

    /// Event that is fired whenever a packet is queued for the specified user.
    SigC::Signal1<void, const std::string&> queue_emptied_event;
private:
    // types
    typedef std::list<judo::XPath::Query*> XPQueryList;
    typedef std::map<Glib::ustring, judo::Element> pqueue;

    // Helper functions
    struct fileToElems : public std::unary_function<judo::Node*, void>
    {
        fileToElems(judo::Element& elem) : _elem(elem) {}
        void operator() (judo::Node* node)
        {
            if (node->getType() == judo::Node::ntElement)
            {
                _elem.appendChild(node);
            }
        }
        judo::Element& _elem;
    };

    struct saveToFile : public std::unary_function<judo::Node*, void>
    {
        saveToFile(std::ofstream& ofs);
        void operator() (const judo::Node* node);
        std::ofstream& _ofs;
    };

    // Vars
    jabberoo::Session& _session;
    std::string _gdir;
};

}; // namespace Gabber

#endif // INCL_GABBER_PACKETQUEUE_HH
