/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Gabber 2
 *  Based on Gabber, Copyright (c) 1999-2002 Dave Smith & Julian Missig
 *  Copyright (c) 2002 Julian Missig
 */

#include "PlainTextView.hh"

#include <message.hh>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/scrollbar.h>
#include <gtkmm/textview.h>
#include <gtkmm/textmark.h>
#include <iostream>

namespace Gabber {

PlainTextView::PlainTextView(Gtk::ScrolledWindow* scroll_win) : 
_scr_win(scroll_win)
{
    _view = new Gtk::TextView();
    _buffer = _view->get_buffer();
    
    // initialize buffer uri/smiley parser
    _bufferparser = new TextBufferParser(_buffer);
    
    // initialize marks:
    _startmark         = _buffer->create_mark("startmark", _buffer->end());
    _bottom            = _buffer->create_mark("bottom", _buffer->end());
    _start_of_last_msg = _buffer->create_mark("start_of_last", _buffer->end());
    
    // Set TextView attributes:
    _view->set_wrap_mode(Gtk::WRAP_WORD);
    _view->set_editable(false);
    _view->set_cursor_visible(false);
    _view->set_left_margin(3); // this way letters won't touch the side edge
    _view->set_right_margin(3);
    _view->property_can_focus().set_value(false);

    // Figure out the line height
    // THEME/FONT SPECIFIC
    Glib::RefPtr<Pango::Layout> playout = Pango::Layout::create(_view->get_pango_context());
    Pango::Rectangle inkrect, logicalrect;
    playout->get_pixel_extents(inkrect, logicalrect);
    
    // Set the pixels_below_lines to be half of the line height.
    _view->set_pixels_below_lines(logicalrect.get_height()/4);
    

    // the Gtk::ScrolledWindow contains our TextView widget:
    _scr_win->add(*_view);

    // store TextView size so we'll know when it changes:
    _view_height = _view->get_height();
    _view_width  = _view->get_width();

    // Remote tag
    Glib::RefPtr<Gtk::TextBuffer::Tag> tag =  _buffer->create_tag("remote");
    tag->property_foreground().set_value("red");

    // Local tag
    tag =  _buffer->create_tag("local");
    tag->property_foreground().set_value("blue");

    // Note tag
    tag = _buffer->create_tag("note");
    tag->property_foreground().set_value("dark green");

    // /me tag
    tag = _buffer->create_tag("action");  

    // Composing tag
    tag = _buffer->create_tag("composing");
    tag->property_style().set_value(Pango::STYLE_ITALIC);

    // Timestamp tag
    tag = _buffer->create_tag("timestamp");
    tag->property_foreground().set_value("dark grey");
    tag->property_justification().set_value(Gtk::JUSTIFY_CENTER);
    int sz = _view->get_pango_context()->get_font_description().get_size();
    tag->property_size().set_value(sz - (sz/4));
}

PlainTextView::~PlainTextView()
{
    delete _bufferparser;
    delete _view;
}

void PlainTextView::append(const jabberoo::Message& m, 
                           const Glib::ustring& nick, bool local_msg)
{
    Glib::ustring from = m.getFrom();
    Glib::ustring tag = "local";
    Gtk::Adjustment* vadj = _scr_win->get_vadjustment();
    bool scroll_to_bottom = is_at_bottom(vadj);

    if (!local_msg) 
    {
        if (_composers.find(from) != _composers.end())
            cancel_composing(from);

        tag = "remote";
    }

    _buffer->move_mark(_start_of_last_msg , _buffer->end());

    const Glib::ustring& body = m.getBody();
    // Check for /me
    if (body.substr(0,3) == "/me" && (body.substr(0,4) == "/me " || body.substr(0,6) == "/me's "))
    {
	    // set the tags for /me
	    std::vector<Glib::ustring> metags(2);
	    metags[0] = "action";
	    metags[1] = tag;

	    // insert the /me text
	    _buffer->insert_with_tags_by_name(_buffer->end(), "* ", metags);
            _buffer->insert_with_tag(_buffer->end(), nick, "action");
	    _buffer->insert_with_tag(_buffer->end(), body.substr(3) + "\n", 
            "action");
    }
    else
    {
        // no /me, insert normally
        _buffer->insert_with_tag(_buffer->end(), "<", tag);
        _buffer->insert(_buffer->end(), nick);
        _buffer->move_mark(_startmark, 
                           _buffer->insert_with_tag(_buffer->end(), "> ", tag));
        _buffer->insert(_buffer->end(), body + "\n");
        _buffer->move_mark(_bottom, _buffer->end());
    
        // URIs and smileys ahoy
        _bufferparser->parse_buffer(_startmark, _bottom);
    }

    _buffer->move_mark(_bottom, _buffer->end());

    if (local_msg || scroll_to_bottom) {
        _view->scroll_to_mark(_bottom, 0, 0, 0);
    }
}

void PlainTextView::append(const Glib::ustring& str)
{
    Gtk::Adjustment* vadj = _scr_win->get_vadjustment();
    bool scroll_to_bottom = is_at_bottom(vadj);

    // set the tags for /me
    std::vector<Glib::ustring> tags(2);
    tags[0] = "action";
    tags[1] = "note";

    // insert the /me text
    _buffer->move_mark(_startmark, 
                       _buffer->insert_with_tags_by_name(_buffer->end(), "* ", tags));
    _buffer->insert_with_tag(_buffer->end(), str + "\n", "action");
    _buffer->move_mark(_bottom, _buffer->end());
    _bufferparser->parse_buffer(_startmark, _bottom);
    _buffer->move_mark(_bottom, _buffer->end());

    if (scroll_to_bottom) {
        _view->scroll_to_mark(_bottom, 0, 0, 0);
    }
}

void PlainTextView::append_timestamp()
{
    Gtk::Adjustment* vadj = _scr_win->get_vadjustment();
    bool scroll_to_bottom = is_at_bottom(vadj);
#ifndef WIN32
    char time_str[8];

    time_t t = time(NULL);
    strftime(time_str, 8, "%H:%M", localtime(&t));
#endif
    // XXX Figure this out for win32
    _buffer->insert_with_tag(_buffer->end(), time_str, "timestamp");
    _buffer->insert(_buffer->end(), "\n");
    _buffer->move_mark(_bottom, _buffer->end());

    if (scroll_to_bottom) {
        _view->scroll_to_mark(_bottom, 0, 0, 0);
    }
}

void PlainTextView::append_error(const Glib::ustring& msg)
{
    Gtk::Adjustment* vadj = _scr_win->get_vadjustment();
    bool scroll_to_bottom = is_at_bottom(vadj);

    _buffer->insert(_buffer->end(), "ERROR: " + msg + "\n");
    _buffer->move_mark(_bottom, _buffer->end());

    if (scroll_to_bottom) {
        _view->scroll_to_mark(_bottom, 0, 0, 0);
    }
}

void PlainTextView::composing(const std::string& jid, const Glib::ustring& nick)
{
    Gtk::Adjustment* vadj = _scr_win->get_vadjustment();
    bool scroll_to_bottom = is_at_bottom(vadj);

    // Alread have it, do something?
    if (_composers.find(jid) != _composers.end())
        return;

    // XXX Use something other than the jid as the key?  That's what the set is
    // for
    _buffer->create_mark(jid + "_start", _buffer->end());
    _buffer->insert_with_tag(_buffer->end(), "(", "composing");
    _buffer->insert_with_tag(_buffer->end(), nick, "composing");
    _buffer->insert_with_tag(_buffer->end(), ") ...", "composing");
    _buffer->insert(_buffer->end(), "\n");
    _buffer->create_mark(jid + "_end", _buffer->end());

    _buffer->move_mark(_bottom, _buffer->end());

    if (scroll_to_bottom) {
        _view->scroll_to_mark(_bottom, 0, 0, 0);
    }

    _composers.insert(jid);
}

void PlainTextView::cancel_composing(const std::string& jid)
{
    Gtk::Adjustment* vadj = _scr_win->get_vadjustment();
    bool scroll_to_bottom = is_at_bottom(vadj);

    mark_map::iterator it = _composers.find(jid);

    // Not in the map, do we error?
    if (it == _composers.end())
        return;

    txt_mark_ref start = _buffer->get_mark(jid + "_start");
    txt_mark_ref end = _buffer->get_mark(jid + "_end");
    if (!start || !end)
    {
        std::cerr << "No marks found for " << jid << std::endl;
        return;
    }
    _buffer->erase(_buffer->get_iter_at_mark(start), 
            _buffer->get_iter_at_mark(end));

    _buffer->delete_mark(start);
    _buffer->delete_mark(end);

    _buffer->move_mark(_bottom, _buffer->end());

    if (scroll_to_bottom) {
        _view->scroll_to_mark(_bottom, 0, 0, 0);
    }

    _composers.erase(it);
}

Gtk::Widget* PlainTextView::get_widget()
{
    return _view;
}

bool PlainTextView::is_at_bottom(const Gtk::Adjustment* vadj) {
    return vadj->get_value() >= vadj->get_upper() - vadj->get_page_size();
}

} // namespace Gabber
