#include "GabberApp.hh"
#include "GabberWidgets.hh"
#include "GabberUtility.hh"
#include "ChatView.hh"
#include "ChatViewManager.hh"
#include "LogManager.hh"
#include "XPathLog.hh"
#include "PlainTextView.hh"
#include "PrettyText.hh"

#include "gtkspell.h"

#include "intl.h"
#include <sigc++/slot.h>

#include <gtkmm/messagedialog.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/stock.h>
#include <gtkmm/button.h>
#include <gtkmm/textview.h>
#include <gtkmm/togglebutton.h>

namespace Gabber {

// -------------------------------------------------------------------
//
// Chat (One-on-One) Message View
//
// -------------------------------------------------------------------

ChatView::ChatView(ChatViewManager& mgr, const judo::Element& msg)
     :BaseGabberWindow("OOOChat_win"), _mgr(mgr),
     _jid(msg.getAttrib("from")), _last_msg_time(time(NULL) - 3),
     _sent_composing(false), _composing_id("")
{
     init(false);// Basic initialization
     _incoming_logger->check(msg);
     on_event_node(msg);
     on_chat_node(msg);
}

ChatView::ChatView(ChatViewManager& mgr, const std::string& jid)
     :BaseGabberWindow("OOOChat_win"), _mgr(mgr),
     _jid(jid), _last_msg_time(time(NULL) - 3), _sent_composing(false), 
     _composing_id("")
{
     init(true);// Basic initialization
}


ChatView::~ChatView()
{
    delete _txtChatview;
    delete _incoming_logger;
    delete _outgoing_logger;
}

void ChatView::close()
{
    if(!_composing_id.empty() && _sent_composing)
    {
        // They deleted, stop the event
        Message m(_jid, "", Message::mtChat);
        judo::Element* x = m.addX("jabber:x:event");
        x->addElement("id", _composing_id);

        G_App.getSession() << m;
        
        _sent_composing = false;
    }

    // Don't get xpaths anymore, that'd be bad
     if (!_jid.empty())
     {
          G_App.getSession().unregisterXPath(_chat_query);
          G_App.getSession().unregisterXPath(_event_query);
          G_App.getSession().unregisterXPath(_presence_query);
          
          // unregister chat
          _mgr.releaseChat(_jid);
     }
     
    // destroy the Gtk::Window
    BaseGabberWindow::close();
}
    
void ChatView::init(bool is_blank)
{
    // Get widgets
    get_widget("AddContact_btn", _btnAddContact);
    get_widget("SendAsMsg_tglbtn", _tglbtnSendAsMsg);
    get_widget("Message_txt", _txtMessage);

    // GtkSpell
    GError *error = NULL;
    char *errortext = NULL;
    if (gtkspell_new_attach(GTK_TEXT_VIEW(_txtMessage->gobj()), NULL, &error) == NULL) {
	 g_print("gtkspell error: %s\n", error->message);
	 errortext = g_strdup_printf("GtkSpell was unable to initialize.\n"
				     "%s", error->message);
	 g_error_free(error);
    }

    _txtMessage->get_buffer()->signal_changed().connect(
        slot(*this, &ChatView::on_txtMessage_changed));

    // Figure out the line height
    // THEME/FONT SPECIFIC
    Glib::RefPtr<Pango::Layout> playout = 
    Pango::Layout::create(_txtMessage->get_pango_context());
    Pango::Rectangle inkrect, logicalrect;
    playout->get_pixel_extents(inkrect, logicalrect);

    // Set the text entry area to be 2 lines tall per default
    _txtMessage->set_size_request(-1, logicalrect.get_height() * 2 + 4);

    Gtk::ScrolledWindow* scroll_win;
    get_widget("Chatview_scroll", scroll_win);
    _txtChatview = new PlainTextView(scroll_win);
    _txtChatview->append_timestamp();
    scroll_win->show_all();

    // Nickname and status display
    if (is_blank)
    {
        // We allow resource selection
        _prettyjid = manage(new PrettyJID(_jid, "", PrettyJID::dtNickRes, 
                                          true));
    }
    else if (_jid.empty())
    {
         _prettyjid = manage(new PrettyJID(_jid, "", PrettyJID::dtNickRes,
                                           true, true));
    }
    else
    {
        _prettyjid = manage(new PrettyJID(_jid, "", PrettyJID::dtNickRes, 
                                          false));
    }
    _prettyjid->show();
    _prettyjid->changed.connect(slot(*this, &ChatView::on_resource_changed));

    Gtk::HBox* hb;
    get_widget("JIDInfo_hbox", hb);
    hb->pack_start(*_prettyjid, true, true, 0);
    _nickname = _prettyjid->get_nickname();

    // Set the nicknames
    _local_nick = G_App.getSession().getUserName();

    // Connect events
    _thisWindow->signal_event().connect(slot(*this, 
        &ChatView::on_window_event));

    // Set the window title
    _thisWindow->set_title(Util::substitute(_("Chat with %s"), _nickname));

    if (!_jid.empty())
    {
        // Register to recieve chat messages from _jid
        _chat_query = G_App.getSession().registerXPath(
            "/message[@from='" + _jid + "'][@type='chat' or @type='error']",
            slot(*this, &ChatView::on_chat_node));
        // Register to recieve chat events addressed from _jid
        _event_query = G_App.getSession().registerXPath(
            "/message[@from='" + _jid + "']/x[@xmlns='jabber:x:event']", 
            slot(*this, &ChatView::on_event_node));

        // Register for presence events
        _presence_query = G_App.getSession().registerXPath(
            "/presence[@from='" + _jid + "']", 
            slot(*this, &ChatView::on_presence_node));

        // Setup the loggers, only packets with bodies? XXX
        _incoming_logger = G_App.getLogManager().create_xpath_log(
            "/message[@type='chat' and @from='" + _jid + "']/body", _jid);
        _outgoing_logger = G_App.getLogManager().create_xpath_log(
            "/message[@type='chat' and @to='" + _jid + "']/body", _jid, false);
    }
    
    // Grab any existing presence
    try 
    {
        const Presence& p = G_App.getSession().presenceDB().findExact(
            _prettyjid->get_full_jid());
        on_presence_node(p.getBaseElement());
    } 
    catch (PresenceDB::XCP_InvalidJID& e) 
    {
        Presence p("", Presence::ptUnavailable, Presence::stOffline, 
            _("No presence has been received."));
        on_presence_node(p.getBaseElement());
    }

    // Display the window
    show();
    _thisWindow->raise();
}

void ChatView::on_Send_clicked()
{
    // if we're selecting a resource
    if (_prettyjid->is_selecting_resource())
    {
	    // Then lock down the resource
	    on_resource_locked();
    }

    // Add a timestamp if the gap has been too large
    // XXX configurable timestamp gap?
    time_t t = time(NULL);
    if ( (t - _last_msg_time) >= 300 )
    {
        _txtChatview->append_timestamp();
    }
    _last_msg_time = t;

    Glib::RefPtr<Gtk::TextBuffer> tb = _txtMessage->get_buffer();
    Glib::ustring body = tb->get_text(tb->begin(), tb->end());
    
    // Bogus
    if (body.empty())
        return;

    // Construct message
    Message m(_jid, body, Message::mtChat);

    // XXX Need disco or something here to check this a better way
    m.requestComposing();
    m.setID(G_App.getSession().getNextID());

    if (!_thread.empty())
	 m.setThread(_thread);

    G_App.getSession() << m; // Send the message

    _txtChatview->cancel_composing("");
    _txtChatview->append(m, _local_nick, true);

    _txtMessage->get_buffer()->set_text(Glib::ustring()); // reset the message area
    _sent_composing = false; // don't send a cancel composing message, because we haven't sent composing
}

bool ChatView::check_close()
{
    time_t now = time(NULL);
    // XXX configurable time somehow?
    if ( (now - _last_received_time) <= 3 )
    {
        Gtk::MessageDialog dlg(*_thisWindow, 
            Util::substitute(_("A message was just received from %s, are you sure you want to close this chat?"), _nickname),
            Gtk::MESSAGE_WARNING, Gtk::BUTTONS_NONE, true);
	dlg.set_title(Util::substitute(_("Close chat with %s"), _nickname));
	dlg.add_button(_("_Don't Close"), Gtk::RESPONSE_NO);
	dlg.add_button(Gtk::Stock::CLOSE, Gtk::RESPONSE_YES);

        int ret = dlg.run();
        dlg.hide();
        if (ret == Gtk::RESPONSE_NO)
            return false;
    }

    return true;
}

bool ChatView::on_delete_event(GdkEventAny* e)
{
    if(!check_close())
        return true;

    BaseGabberWindow::on_delete_event(e);

    return false;
}

bool ChatView::on_window_event(GdkEvent* ev) 
{
    if (ev->type != GDK_KEY_PRESS)
    return false;

    GdkEventKey* e = (GdkEventKey*)ev;

    // escape closes our window
    if (e->keyval == GDK_Escape)
    {
        if (check_close())
            close();
        return false;
    }

    // If they pressed the Keypad enter, make it act like a normal enter
    if (e->keyval == GDK_KP_Enter)
        e->keyval = GDK_Return;

    if (e->keyval == GDK_Return)
    {
        //enter a newline if shift-return is used
        if (e->state & GDK_SHIFT_MASK)
        {
            //unset the shift bit. shift-return seems to have a special meaning for the widget
            e->state ^= GDK_SHIFT_MASK;
            return false;
        }

        // Ctrl-Return sends the message so we don't need to check for it  
        on_Send_clicked();
        return true;
    }

    return false;
}

void ChatView::on_txtMessage_changed()
{
    // Normal typing, send the appropiate composing msg
    int chars = _txtMessage->get_buffer()->get_char_count();
    if (chars == 0)
    {
        if(!_composing_id.empty() && _sent_composing)
        {
            // They deleted, stop the event
            Message m(_jid, "", Message::mtChat);
            judo::Element* x = m.addX("jabber:x:event");
            x->addElement("id", _composing_id);

            G_App.getSession() << m;
            
            _sent_composing = false;
        }

        _txtChatview->cancel_composing("");
    }
    else if (chars > 0)
    {
        if (!_composing_id.empty() && !_sent_composing)
        {
            // Send a fresh composing
            Message m(_jid, "", Message::mtChat);
            judo::Element* x = m.addX("jabber:x:event");
            x->addElement("composing");
            x->addElement("id", _composing_id);

            G_App.getSession() << m;

            _sent_composing = true;
        }

        _txtChatview->composing("", _local_nick);
    }
}

void ChatView::on_chat_node(const judo::Element& e)
{
    // XXX If this window is not visible we copy into _pending_msgs
    if (e.getAttrib("type") == "error")
    {
        // XXX Do more?!
        printf("ERROR: %s\n", e.toString().c_str());
        const judo::Element* err = e.findElement("error");
        _txtChatview->append_error(err->getCDATA());
    }
    else
    {
        display(Message(e));
    }
}

void ChatView::on_event_node(const judo::Element& e)
{
    // Avoid errors
    if (e.getAttrib("type") == "error")
        return;

    Message m(e);

    const judo::Element* x = m.findX("jabber:x:event");
    if (!x)
        return;
    const judo::Element* delivered = x->findElement("delivered");
    const judo::Element* composing = x->findElement("composing");
    const judo::Element* id = x->findElement("id");

    if (composing && id)
    {
        // If we have both of these they are typing
        _txtChatview->composing(e.getAttrib("from"), _nickname);
    }
    else if (!composing && id)
    {
        // cancel the events
        _txtChatview->cancel_composing(e.getAttrib("from"));
    }
    else if (composing && !id)
    {
        _composing_id = m.getID();
    }
    else if (delivered)
    {
        // Always handle these, if they are in there
        G_App.getSession() << m.delivered();
    }
}

void ChatView::on_presence_node(const judo::Element& e)
{
     Presence p(e);
     
     Gtk::HBox* hb;
     get_widget("Show_hbox", hb);

     Gtk::Label* l;
     get_widget("Status_lbl", l);

     // Show the extra bar if they are not available or chatty
     if (p.getShow() == Presence::stOnline || p.getShow() == Presence::stChat)
     {
	  hb->hide();
     }
     else
     {
	  hb->show();
	  // If there is an explanation to go along with the short presence, display it properly
	  if (!p.getStatus().empty())
	  {
	       l->set_label("<b>" + Util::getShowName(p.getShow()) + "</b>: " + p.getStatus());
	  }
	  else
	  {
	       l->set_label("<b>" + Util::getShowName(p.getShow()) + "</b>");
	  }
     }

     // Prevent sending composing events anymore if user goes offline
     if (p.getShow() == Presence::stOffline)
     {
          _composing_id = Glib::ustring();
     }
}

void ChatView::on_resource_changed()
{
     G_App.getSession().unregisterXPath(_presence_query);

     // Register for presence events
     _presence_query = G_App.getSession().registerXPath(
	  "/presence[@from='" + _jid + "']", 
	  slot(*this, &ChatView::on_presence_node));
		
     // Grab any existing presence
     try {
	  const Presence& p = G_App.getSession().presenceDB().findExact(_prettyjid->get_full_jid());
	  on_presence_node(p.getBaseElement());
     } catch (PresenceDB::XCP_InvalidJID& e) {
	  Presence p("", Presence::ptUnavailable, Presence::stOffline, _("No presence has been received."));
	  on_presence_node(p.getBaseElement());
     }
}

void ChatView::on_resource_locked()
{
     // hide the resource selection now
     _prettyjid->hide_resource_select();

     // unregister the general jid
     if (!_jid.empty())
     {
          G_App.getSession().unregisterXPath(_chat_query);
          G_App.getSession().unregisterXPath(_event_query);
          _mgr.releaseChat(_jid);
     }
     else
     {
          _jid = _prettyjid->get_full_jid();
          // Setup the loggers, only packets with bodies? XXX
          _incoming_logger = G_App.getLogManager().create_xpath_log(
               "/message[@type='chat' and @from='" + _jid + "']/body", _jid);
          _outgoing_logger = G_App.getLogManager().create_xpath_log(
               "/message[@type='chat' and @to='" + _jid + "']/body", _jid, false);
     }
     
     // get the new jid
     _jid = _prettyjid->get_full_jid();

     // if there's no resource, we probably want to do something
     // like catch whatever message is sent back.. ? XXX
     if (JID::getResource(_jid).empty())
     {
	  cout << "empty resource selected" << endl;
     }
     
     // register the specific jid
     _mgr.register_chat(_jid, this);
     // Register to recieve chat messages from _jid
     _chat_query = G_App.getSession().registerXPath(
	  "/message[@type='chat' and @from='" + _jid + "']",
	  slot(*this, &ChatView::on_chat_node));
     // Register to recieve chat events addressed from _jid
     _event_query = G_App.getSession().registerXPath(
	  "/message[@from='" + _jid + 
	  "']/x[@xmlns='jabber:x:event']", slot(*this, 
						&ChatView::on_event_node));
}

void ChatView::display(const Message& m)
{
    // Track the last time we got a message to warn on fast close, we also put
    // in a timestamp if there has been a gap of 5 minutes or more
    time_t t = time(NULL);
    if ( (t - _last_msg_time) >= 300 )
    {
        _txtChatview->append_timestamp();
    }
    _last_received_time = t;
    _last_msg_time = t;

    // bail on useless nodes
    if (m.getBody().empty()) return;

    // Display it
    _txtChatview->append(m, _nickname);

    // Save the thread!
    _thread = m.getThread();

    // Say we were displayed if they want that
    judo::Element* x = m.findX("jabber:x:event");
    if (x && x->findElement("displayed"))
        G_App.getSession() << m.displayed();
}

} // namespace Gabber
